#include<stdio.h>
#include<stdlib.h>
#include<openacc.h>
#include<string.h>
#include<math.h>
#include<mpi.h>
#include<stdbool.h>

double add_vector(int n, bool if_acc);

double add_vector(int n, bool if_acc)
{
  // Input vectors
  double *restrict a;
  double *restrict b;
  // Output vector
  double *restrict c;
 
  // Size, in bytes, of each vector
  size_t bytes = n*sizeof(double);
 
  // Allocate memory for each vector
  a = (double*)malloc(bytes);
  b = (double*)malloc(bytes);
  c = (double*)malloc(bytes);
 
  // Initialize content of input vectors, vector a[i] = sin(i)^2 vector b[i] = cos(i)^2
  int i;
  for(i=0; i<n; i++) {
    a[i] = i; 
    b[i] = i; 
    c[i] = 0;
  }  
 
  // sum component wise and save result into vector c
#pragma acc enter data create(a[0:bytes], b[0:bytes], c[0:bytes]) if(if_acc)
#pragma acc kernels copyin(a[0:n],b[0:n]), copyout(c[0:n]) if(if_acc)
  for(i=0; i<n; i++) {
    c[i] = a[i] + b[i];
  }
 
  // Sum up vector c and print result divided by n, this should equal 1 within error
  double sum = 0.0;
  for(i=0; i<n; i++) {
    sum += c[i];
  }
  sum = sum/n;
  return sum;
 
}

int main(int argc, char **argv)
{
  int ThisTask;

  bool if_acc;

  MPI_Init(&argc, &argv);

  MPI_Comm_rank(MPI_COMM_WORLD, &ThisTask);

  char *device=getenv("ACC_DEVICE_TYPE");

  //  printf("device type %s\n", device);
 
  if(strcmp(device, "acc_device_host") == 0)
    {
      acc_init(acc_device_host);
      acc_set_device_type(acc_device_host);
      printf("the value of device type is %d\n", acc_device_host); 
      const int num_devices = acc_get_num_devices(acc_device_host); 
      if_acc = true;
      double sum = add_vector(100, if_acc);
      printf("sum on cpu %f, num_devices \n", sum, num_devices);
    }
  else
    {
      if(strcmp(device, "acc_device_nvidia") == 0)
	{
	  acc_init(acc_device_nvidia);
	  acc_set_device_type(acc_device_nvidia);
	  const int num_devices = acc_get_num_devices(acc_device_nvidia);
	  if_acc = true;
	  double sum = add_vector(1000000, if_acc);
	  printf("sum on gpu %f, num devices \n", sum, num_devices);
	}
      else
	{
	  if_acc = false;
	  double sum = 0;
	  printf("before add\n");
	  sum = add_vector(10, if_acc);
	  printf("sum on pure cpu %f \n", sum);
	}
    }

  return 0;
}
